@@ -6,18 +6,38 @@ import numpy as np
 import pandas
 import logging
 import uuid
+import re
+import warnings
 from pathlib import Path
 from .settings import config
 from .declare import declare, alter
+from .condition import make_condition
 from .expression import QueryExpression
 from . import blob
-from .utils import user_choice, OrderedDict
+from .utils import user_choice
 from .heading import Heading
-from .errors import DuplicateError, AccessError, DataJointError, UnknownAttributeError
+from .errors import (DuplicateError, AccessError, DataJointError, UnknownAttributeError,
+                     IntegrityError)
 from .version import __version__ as version
 
 logger = logging.getLogger(__name__)
 
+foreign_key_error_regexp = re.compile(
+    r"[\w\s:]*\((?P<child>`[^`]+`.`[^`]+`), "
+    r"CONSTRAINT (?P<name>`[^`]+`) "
+    r"(FOREIGN KEY \((?P<fk_attrs>[^)]+)\) "
+    r"REFERENCES (?P<parent>`[^`]+`(\.`[^`]+`)?) \((?P<pk_attrs>[^)]+)\)[\s\w]+\))?")
+
+constraint_info_query = ' '.join("""
+    SELECT
+        COLUMN_NAME as fk_attrs,
+        CONCAT('`', REFERENCED_TABLE_SCHEMA, '`.`', REFERENCED_TABLE_NAME, '`') as parent,
+        REFERENCED_COLUMN_NAME as pk_attrs
+    FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
+    WHERE
+        CONSTRAINT_NAME = %s AND TABLE_SCHEMA = %s AND TABLE_NAME = %s;
+    """.split())
+
 
 class _RenameMap(tuple):
     """ for internal use """
@@ -31,23 +51,21 @@ class Table(QueryExpression):
     table name, database, and definition.
     A Relation implements insert and delete methods in addition to inherited relational operators.
     """
-    _heading = None
+
+    _table_name = None  # must be defined in subclass
+    _log_ = None  # placeholder for the Log table object
+
+    # These properties must be set by the schema decorator (schemas.py) at class level or by FreeTable at instance level
     database = None
-    _log_ = None
     declaration_context = None
 
-    # -------------- required by QueryExpression ----------------- #
     @property
-    def heading(self):
-        """
-        :return: table heading. If the table is not declared, attempts to declare it first.
-        """
-        if self._heading is None:
-            self._heading = Heading()  # instance-level heading
-        if not self._heading and self.connection is not None:  # lazy loading of heading
-            self._heading.init_from_database(
-                self.connection, self.database, self.table_name, self.declaration_context)
-        return self._heading
+    def table_name(self):
+        return self._table_name
+
+    @property
+    def definition(self):
+        raise NotImplementedError('Subclasses of Table must implement the `definition` property')
 
     def declare(self, context=None):
         """
@@ -76,8 +94,8 @@ class Table(QueryExpression):
         Alter the table definition from self.definition
         """
         if self.connection.in_transaction:
-            raise DataJointError('Cannot update table declaration inside a transaction, '
-                                 'e.g. from inside a populate/make call')
+            raise DataJointError(
+                'Cannot update table declaration inside a transaction, e.g. from inside a populate/make call')
         if context is None:
             frame = inspect.currentframe().f_back
             context = dict(frame.f_globals, **frame.f_locals)
@@ -99,11 +117,11 @@ class Table(QueryExpression):
                     # skip if no create privilege
                     pass
                 else:
+                    self.__class__._heading = Heading(table_info=self.heading.table_info)  # reset heading
                     if prompt:
                         print('Table altered')
                     self._log('Altered ' + self.full_table_name)
 
-    @property
     def from_clause(self):
         """
         :return: the FROM clause of SQL SELECT statements.
@@ -205,6 +223,40 @@ class Table(QueryExpression):
     def external(self):
         return self.connection.schemas[self.database].external
 
+    def update1(self, row):
+        """
+        update1 updates one existing entry in the table.
+        Caution: Updates are not part of the DataJoint data manipulation model. For strict data integrity,
+        use delete and insert.
+        :param row: a dict containing the primary key and the attributes to update.
+        Setting an attribute value to None will reset it to the default value (if any)
+        The primary key attributes must always be provided.
+        Examples:
+        >>> table.update1({'id': 1, 'value': 3})  # update value in record with id=1
+        >>> table.update1({'id': 1, 'value': None})  # reset value to default
+        """
+        # argument validations
+        if not isinstance(row, collections.abc.Mapping):
+            raise DataJointError('The argument of update1 must be dict-like.')
+        if not set(row).issuperset(self.primary_key):
+            raise DataJointError('The argument of update1 must supply all primary key values.')
+        try:
+            raise DataJointError('Attribute `%s` not found.' % next(k for k in row if k not in self.heading.names))
+        except StopIteration:
+            pass  # ok
+        if len(self.restriction):
+            raise DataJointError('Update cannot be applied to a restricted table.')
+        key = {k: row[k] for k in self.primary_key}
+        if len(self & key) != 1:
+            raise DataJointError('Update entry must exist.')
+        # UPDATE query
+        row = [self.__make_placeholder(k, v) for k, v in row.items() if k not in self.primary_key]
+        query = "UPDATE {table} SET {assignments} WHERE {where}".format(
+            table=self.full_table_name,
+            assignments=",".join('`%s`=%s' % r[:2] for r in row),
+            where=make_condition(self, key, set()))
+        self.connection.query(query, args=list(r[2] for r in row if r[2] is not None))
+
     def insert1(self, row, **kwargs):
         """
         Insert one data record or one Mapping (like a dict).
@@ -216,7 +268,6 @@ class Table(QueryExpression):
     def insert(self, rows, replace=False, skip_duplicates=False, ignore_extra_fields=False, allow_direct_insert=None):
         """
         Insert a collection of rows.
-
         :param rows: An iterable where an element is a numpy record, a dict-like object, a pandas.DataFrame, a sequence,
             or a query expression with the same heading as table self.
         :param replace: If True, replaces the existing tuple.
@@ -224,13 +275,11 @@ class Table(QueryExpression):
         :param ignore_extra_fields: If False, fields that are not in the heading raise error.
         :param allow_direct_insert: applies only in auto-populated tables.
                                     If False (default), insert are allowed only from inside the make callback.
-
         Example::
         >>> relation.insert([
         >>>     dict(subject_id=7, species="mouse", date_of_birth="2014-09-01"),
         >>>     dict(subject_id=8, species="mouse", date_of_birth="2014-09-02")])
         """
-
         if isinstance(rows, pandas.DataFrame):
             # drop 'extra' synthetic index for 1-field index case -
             # frames with more advanced indices should be prepared by user.
@@ -244,7 +293,6 @@ class Table(QueryExpression):
                 'Inserts into an auto-populated table can only done inside its make method during a populate call.'
                 ' To override, set keyword argument allow_direct_insert=True.')
 
-        heading = self.heading
         if inspect.isclass(rows) and issubclass(rows, QueryExpression):   # instantiate if a class
             rows = rows()
         if isinstance(rows, QueryExpression):
@@ -253,235 +301,144 @@ class Table(QueryExpression):
                 try:
                     raise DataJointError(
                         "Attribute %s not found. To ignore extra attributes in insert, set ignore_extra_fields=True." %
-                        next(name for name in rows.heading if name not in heading))
+                        next(name for name in rows.heading if name not in self.heading))
                 except StopIteration:
                     pass
-            fields = list(name for name in rows.heading if name in heading)
+            fields = list(name for name in rows.heading if name in self.heading)
             query = '{command} INTO {table} ({fields}) {select}{duplicate}'.format(
                 command='REPLACE' if replace else 'INSERT',
                 fields='`' + '`,`'.join(fields) + '`',
                 table=self.full_table_name,
-                select=rows.make_sql(select_fields=fields),
+                select=rows.make_sql(fields),
                 duplicate=(' ON DUPLICATE KEY UPDATE `{pk}`={table}.`{pk}`'.format(
                     table=self.full_table_name, pk=self.primary_key[0])
                            if skip_duplicates else ''))
             self.connection.query(query)
             return
 
-        if heading.attributes is None:
-            logger.warning('Could not access table {table}'.format(table=self.full_table_name))
-            return
-
-        field_list = None  # ensures that all rows have the same attributes in the same order as the first row.
-
-        def make_row_to_insert(row):
-            """
-            :param row:  A tuple to insert
-            :return: a dict with fields 'names', 'placeholders', 'values'
-            """
-            def make_placeholder(name, value):
-                """
-                For a given attribute `name` with `value`, return its processed value or value placeholder
-                as a string to be included in the query and the value, if any, to be submitted for
-                processing by mysql API.
-                :param name:  name of attribute to be inserted
-                :param value: value of attribute to be inserted
-                """
-                if ignore_extra_fields and name not in heading:
-                    return None
-                attr = heading[name]
-                if attr.adapter:
-                    value = attr.adapter.put(value)
-                if value is None or (attr.numeric and (value == '' or np.isnan(np.float(value)))):
-                    # set default value
-                    placeholder, value = 'DEFAULT', None
-                else:  # not NULL
-                    placeholder = '%s'
-                    if attr.uuid:
-                        if not isinstance(value, uuid.UUID):
-                            try:
-                                value = uuid.UUID(value)
-                            except (AttributeError, ValueError):
-                                raise DataJointError(
-                                    'badly formed UUID value {v} for attribute `{n}`'.format(v=value, n=name)) from None
-                        value = value.bytes
-                    elif attr.is_blob:
-                        value = blob.pack(value)
-                        value = self.external[attr.store].put(value).bytes if attr.is_external else value
-                    elif attr.is_attachment:
-                        attachment_path = Path(value)
-                        if attr.is_external:
-                            # value is hash of contents
-                            value = self.external[attr.store].upload_attachment(attachment_path).bytes
-                        else:
-                            # value is filename + contents
-                            value = str.encode(attachment_path.name) + b'\0' + attachment_path.read_bytes()
-                    elif attr.is_filepath:
-                        value = self.external[attr.store].upload_filepath(value).bytes
-                    elif attr.numeric:
-                        value = str(int(value) if isinstance(value, bool) else value)
-                return name, placeholder, value
-
-            def check_fields(fields):
-                """
-                Validates that all items in `fields` are valid attributes in the heading
-                :param fields: field names of a tuple
-                """
-                if field_list is None:
-                    if not ignore_extra_fields:
-                        for field in fields:
-                            if field not in heading:
-                                raise KeyError(u'`{0:s}` is not in the table heading'.format(field))
-                elif set(field_list) != set(fields).intersection(heading.names):
-                    raise DataJointError('Attempt to insert rows with different fields')
-
-            if isinstance(row, np.void):  # np.array
-                check_fields(row.dtype.fields)
-                attributes = [make_placeholder(name, row[name])
-                              for name in heading if name in row.dtype.fields]
-            elif isinstance(row, collections.abc.Mapping):  # dict-based
-                check_fields(row)
-                attributes = [make_placeholder(name, row[name]) for name in heading if name in row]
-            else:  # positional
-                try:
-                    if len(row) != len(heading):
-                        raise DataJointError(
-                            'Invalid insert argument. Incorrect number of attributes: '
-                            '{given} given; {expected} expected'.format(
-                                given=len(row), expected=len(heading)))
-                except TypeError:
-                    raise DataJointError('Datatype %s cannot be inserted' % type(row))
-                else:
-                    attributes = [make_placeholder(name, value) for name, value in zip(heading, row)]
-            if ignore_extra_fields:
-                attributes = [a for a in attributes if a is not None]
-
-            assert len(attributes), 'Empty tuple'
-            row_to_insert = dict(zip(('names', 'placeholders', 'values'), zip(*attributes)))
-            nonlocal field_list
-            if field_list is None:
-                # first row sets the composition of the field list
-                field_list = row_to_insert['names']
-            else:
-                #  reorder attributes in row_to_insert to match field_list
-                order = list(row_to_insert['names'].index(field) for field in field_list)
-                row_to_insert['names'] = list(row_to_insert['names'][i] for i in order)
-                row_to_insert['placeholders'] = list(row_to_insert['placeholders'][i] for i in order)
-                row_to_insert['values'] = list(row_to_insert['values'][i] for i in order)
-
-            return row_to_insert
-
-        rows = list(make_row_to_insert(row) for row in rows)
+        field_list = []  # collects the field list from first row (passed by reference)
+        rows = list(self.__make_row_to_insert(row, field_list, ignore_extra_fields) for row in rows)
         if rows:
             try:
                 query = "{command} INTO {destination}(`{fields}`) VALUES {placeholders}{duplicate}".format(
                     command='REPLACE' if replace else 'INSERT',
-                    destination=self.from_clause,
+                    destination=self.from_clause(),
                     fields='`,`'.join(field_list),
                     placeholders=','.join('(' + ','.join(row['placeholders']) + ')' for row in rows),
                     duplicate=(' ON DUPLICATE KEY UPDATE `{pk}`=`{pk}`'.format(pk=self.primary_key[0])
                                if skip_duplicates else ''))
                 self.connection.query(query, args=list(
-                    itertools.chain.from_iterable((v for v in r['values'] if v is not None) for r in rows)))
+                    itertools.chain.from_iterable(
+                        (v for v in r['values'] if v is not None) for r in rows)))
             except UnknownAttributeError as err:
-                raise err.suggest('To ignore extra fields in insert, set ignore_extra_fields=True') from None
+                raise err.suggest(
+                    'To ignore extra fields in insert, set ignore_extra_fields=True')
             except DuplicateError as err:
-                raise err.suggest('To ignore duplicate entries in insert, set skip_duplicates=True') from None
+                raise err.suggest(
+                    'To ignore duplicate entries in insert, set skip_duplicates=True')
 
     def delete_quick(self, get_count=False):
         """
         Deletes the table without cascading and without user prompt.
         If this table has populated dependent tables, this will fail.
         """
-        query = 'DELETE FROM ' + self.full_table_name + self.where_clause
+        query = 'DELETE FROM ' + self.full_table_name + self.where_clause()
         self.connection.query(query)
         count = self.connection.query("SELECT ROW_COUNT()").fetchone()[0] if get_count else None
         self._log(query[:255])
         return count
 
-    def delete(self, verbose=True):
+    def _delete_cascade(self):
+        """service function to perform cascading deletes recursively."""
+        max_attempts = 50
+        delete_count = 0
+        for _ in range(max_attempts):
+            try:
+                delete_count += self.delete_quick(get_count=True)
+            except IntegrityError as error:
+                match = foreign_key_error_regexp.match(error.args[0]).groupdict()
+                if "`.`" not in match['child']:  # if schema name missing, use self
+                    match['child'] = '{}.{}'.format(self.full_table_name.split(".")[0],
+                                                    match['child'])
+                if match['pk_attrs'] is not None:  # fully matched, adjusting the keys
+                    match['fk_attrs'] = [k.strip('`') for k in match['fk_attrs'].split(',')]
+                    match['pk_attrs'] = [k.strip('`') for k in match['pk_attrs'].split(',')]
+                else:  # only partially matched, querying with constraint to determine keys
+                    match['fk_attrs'], match['parent'], match['pk_attrs'] = list(map(
+                        list, zip(*self.connection.query(constraint_info_query, args=(
+                            match['name'].strip('`'),
+                            *[_.strip('`') for _ in match['child'].split('`.`')]
+                            )).fetchall())))
+                    match['parent'] = match['parent'][0]
+                # restrict child by self if
+                # 1. if self's restriction attributes are not in child's primary key
+                # 2. if child renames any attributes
+                # otherwise restrict child by self's restriction.
+                child = FreeTable(self.connection, match['child'])
+                if set(self.restriction_attributes) <= set(child.primary_key) and \
+                        match['fk_attrs'] == match['pk_attrs']:
+                    child._restriction = self._restriction
+                elif match['fk_attrs'] != match['pk_attrs']:
+                    child &= self.proj(**dict(zip(match['fk_attrs'],
+                                                  match['pk_attrs'])))
+                else:
+                    child &= self.proj()
+                delete_count += child._delete_cascade()
+            else:
+                print("Deleting {count} rows from {table}".format(
+                    count=delete_count, table=self.full_table_name))
+                break
+        else:
+            raise DataJointError('Exceeded maximum number of delete attempts.')
+        return delete_count
+
+    def delete(self, transaction=True, safemode=None):
         """
         Deletes the contents of the table and its dependent tables, recursively.
-        User is prompted for confirmation if config['safemode'] is set to True.
+
+        :param transaction: if True, use the entire delete becomes an atomic transaction.
+        :param safemode: If True, prohibit nested transactions and prompt to confirm. Default
+            is dj.config['safemode'].
         """
-        conn = self.connection
-        already_in_transaction = conn.in_transaction
-        safe = config['safemode']
-        if already_in_transaction and safe:
-            raise DataJointError('Cannot delete within a transaction in safemode. '
-                                 'Set dj.config["safemode"] = False or complete the ongoing transaction first.')
-        graph = conn.dependencies
-        graph.load()
-        delete_list = OrderedDict(
-            (name, _RenameMap(next(iter(graph.parents(name).items()))) if name.isdigit() else FreeTable(conn, name))
-            for name in graph.descendants(self.full_table_name))
+        safemode = config['safemode'] if safemode is None else safemode
 
-        # construct restrictions for each relation
-        restrict_by_me = set()
-        # restrictions: Or-Lists of restriction conditions for each table.
-        # Uncharacteristically of Or-Lists, an empty entry denotes "delete everything".
-        restrictions = collections.defaultdict(list)
-        # restrict by self
-        if self.restriction:
-            restrict_by_me.add(self.full_table_name)
-            restrictions[self.full_table_name].append(self.restriction)  # copy own restrictions
-        # restrict by renamed nodes
-        restrict_by_me.update(table for table in delete_list if table.isdigit())  # restrict by all renamed nodes
-        # restrict by secondary dependencies
-        for table in delete_list:
-            restrict_by_me.update(graph.children(table, primary=False))   # restrict by any non-primary dependents
-
-        # compile restriction lists
-        for name, table in delete_list.items():
-            for dep in graph.children(name):
-                # if restrict by me, then restrict by the entire relation otherwise copy restrictions
-                restrictions[dep].extend([table] if name in restrict_by_me else restrictions[name])
-
-        # apply restrictions
-        for name, table in delete_list.items():
-            if not name.isdigit() and restrictions[name]:  # do not restrict by an empty list
-                table.restrict([
-                    r.proj() if isinstance(r, FreeTable) else (
-                        delete_list[r[0]].proj(**{a: b for a, b in r[1]['attr_map'].items()})
-                        if isinstance(r, _RenameMap) else r)
-                    for r in restrictions[name]])
-        if safe:
-            print('About to delete:')
-
-        if not already_in_transaction:
-            conn.start_transaction()
-        total = 0
-        try:
-            for name, table in reversed(list(delete_list.items())):
-                if not name.isdigit():
-                    count = table.delete_quick(get_count=True)
-                    total += count
-                    if (verbose or safe) and count:
-                        print('{table}: {count} items'.format(table=name, count=count))
-        except:
-            # Delete failed, perhaps due to insufficient privileges. Cancel transaction.
-            if not already_in_transaction:
-                conn.cancel_transaction()
-            raise
-        else:
-            assert not (already_in_transaction and safe)
-            if not total:
-                print('Nothing to delete')
-                if not already_in_transaction:
-                    conn.cancel_transaction()
+        # Start transaction
+        if transaction:
+            if not self.connection.in_transaction:
+                self.connection.start_transaction()
             else:
-                if already_in_transaction:
-                    if verbose:
-                        print('The delete is pending within the ongoing transaction.')
+                if not safemode:
+                    transaction = False
                 else:
-                    if not safe or user_choice("Proceed?", default='no') == 'yes':
-                        conn.commit_transaction()
-                        if verbose or safe:
-                            print('Committed.')
-                    else:
-                        conn.cancel_transaction()
-                        if verbose or safe:
-                            print('Cancelled deletes.')
+                    raise DataJointError(
+                        "Delete cannot use a transaction within an ongoing transaction. "
+                        "Set transaction=False or safemode=False).")
+
+        # Cascading delete
+        try:
+            delete_count = self._delete_cascade()
+        except:
+            if transaction:
+                self.connection.cancel_transaction()
+            raise
+
+        # Confirm and commit
+        if delete_count == 0:
+            if safemode:
+                print('Nothing to delete.')
+            if transaction:
+                self.connection.cancel_transaction()
+        else:
+            if not safemode or user_choice("Commit deletes?", default='no') == 'yes':
+                if transaction:
+                    self.connection.commit_transaction()
+                if safemode:
+                    print('Deletes committed.')
+            else:
+                if transaction:
+                    self.connection.cancel_transaction()
+                if safemode:
+                    print('Deletes cancelled')
 
     def drop_quick(self):
         """
@@ -542,8 +499,8 @@ class Table(QueryExpression):
             self.connection.dependencies.load()
         parents = self.parents(foreign_key_info=True)
         in_key = True
-        definition = ('# ' + self.heading.table_info['comment'] + '\n'
-                      if self.heading.table_info['comment'] else '')
+        definition = ('# ' + self.heading.table_status['comment'] + '\n'
+                      if self.heading.table_status['comment'] else '')
         attributes_thus_far = set()
         attributes_declared = set()
         indexes = self.heading.indexes.copy()
@@ -597,6 +554,8 @@ class Table(QueryExpression):
 
     def _update(self, attrname, value=None):
         """
+            This is a deprecated function to be removed in datajoint 0.14. Use .update1 instead.
+
             Updates a field in an existing tuple. This is not a datajoyous operation and should not be used
             routinely. Relational database maintain referential integrity on the level of a tuple. Therefore,
             the UPDATE operator can violate referential integrity. The datajoyous way to update information is
@@ -607,9 +566,12 @@ class Table(QueryExpression):
                2. the update attribute must not be in primary key
 
             Example:
-            >>> (v2p.Mice() & key).update('mouse_dob', '2011-01-01')
-            >>> (v2p.Mice() & key).update( 'lens')   # set the value to NULL
+            >>> (v2p.Mice() & key)._update('mouse_dob', '2011-01-01')
+            >>> (v2p.Mice() & key)._update( 'lens')   # set the value to NULL
         """
+        warnings.warn(
+            '`_update` is a deprecated function to be removed in datajoint 0.14. '
+            'Use `.update1` instead.')
         if len(self) != 1:
             raise DataJointError('Update is only allowed on one tuple at a time')
         if attrname not in self.heading:
@@ -632,12 +594,113 @@ class Table(QueryExpression):
         else:
             placeholder = '%s' if value is not None else 'NULL'
         command = "UPDATE {full_table_name} SET `{attrname}`={placeholder} {where_clause}".format(
-            full_table_name=self.from_clause,
+            full_table_name=self.from_clause(),
             attrname=attrname,
             placeholder=placeholder,
-            where_clause=self.where_clause)
+            where_clause=self.where_clause())
         self.connection.query(command, args=(value, ) if value is not None else ())
 
+    # --- private helper functions ----
+    def __make_placeholder(self, name, value, ignore_extra_fields=False):
+        """
+        For a given attribute `name` with `value`, return its processed value or value placeholder
+        as a string to be included in the query and the value, if any, to be submitted for
+        processing by mysql API.
+        :param name:  name of attribute to be inserted
+        :param value: value of attribute to be inserted
+        """
+        if ignore_extra_fields and name not in self.heading:
+            return None
+        attr = self.heading[name]
+        if attr.adapter:
+            value = attr.adapter.put(value)
+        if value is None or (attr.numeric and (value == '' or np.isnan(np.float(value)))):
+            # set default value
+            placeholder, value = 'DEFAULT', None
+        else:  # not NULL
+            placeholder = '%s'
+            if attr.uuid:
+                if not isinstance(value, uuid.UUID):
+                    try:
+                        value = uuid.UUID(value)
+                    except (AttributeError, ValueError):
+                        raise DataJointError(
+                            'badly formed UUID value {v} for attribute `{n}`'.format(v=value,
+                                                                                     n=name))
+                value = value.bytes
+            elif attr.is_blob:
+                value = blob.pack(value)
+                value = self.external[attr.store].put(value).bytes if attr.is_external else value
+            elif attr.is_attachment:
+                attachment_path = Path(value)
+                if attr.is_external:
+                    # value is hash of contents
+                    value = self.external[attr.store].upload_attachment(attachment_path).bytes
+                else:
+                    # value is filename + contents
+                    value = str.encode(attachment_path.name) + b'\0' + attachment_path.read_bytes()
+            elif attr.is_filepath:
+                value = self.external[attr.store].upload_filepath(value).bytes
+            elif attr.numeric:
+                value = str(int(value) if isinstance(value, bool) else value)
+        return name, placeholder, value
+
+    def __make_row_to_insert(self, row, field_list, ignore_extra_fields):
+        """
+        Helper function for insert and update
+        :param row:  A tuple to insert
+        :return: a dict with fields 'names', 'placeholders', 'values'
+        """
+
+        def check_fields(fields):
+            """
+            Validates that all items in `fields` are valid attributes in the heading
+            :param fields: field names of a tuple
+            """
+            if not field_list:
+                if not ignore_extra_fields:
+                    for field in fields:
+                        if field not in self.heading:
+                            raise KeyError(u'`{0:s}` is not in the table heading'.format(field))
+            elif set(field_list) != set(fields).intersection(self.heading.names):
+                raise DataJointError('Attempt to insert rows with different fields')
+
+        if isinstance(row, np.void):  # np.array
+            check_fields(row.dtype.fields)
+            attributes = [self.__make_placeholder(name, row[name], ignore_extra_fields)
+                          for name in self.heading if name in row.dtype.fields]
+        elif isinstance(row, collections.abc.Mapping):  # dict-based
+            check_fields(row)
+            attributes = [self.__make_placeholder(name, row[name], ignore_extra_fields)
+                          for name in self.heading if name in row]
+        else:  # positional
+            try:
+                if len(row) != len(self.heading):
+                    raise DataJointError(
+                        'Invalid insert argument. Incorrect number of attributes: '
+                        '{given} given; {expected} expected'.format(
+                            given=len(row), expected=len(self.heading)))
+            except TypeError:
+                raise DataJointError('Datatype %s cannot be inserted' % type(row))
+            else:
+                attributes = [self.__make_placeholder(name, value, ignore_extra_fields)
+                              for name, value in zip(self.heading, row)]
+        if ignore_extra_fields:
+            attributes = [a for a in attributes if a is not None]
+
+        assert len(attributes), 'Empty tuple'
+        row_to_insert = dict(zip(('names', 'placeholders', 'values'), zip(*attributes)))
+        if not field_list:
+            # first row sets the composition of the field list
+            field_list.extend(row_to_insert['names'])
+        else:
+            #  reorder attributes in row_to_insert to match field_list
+            order = list(row_to_insert['names'].index(field) for field in field_list)
+            row_to_insert['names'] = list(row_to_insert['names'][i] for i in order)
+            row_to_insert['placeholders'] = list(row_to_insert['placeholders'][i] for i in order)
+            row_to_insert['values'] = list(row_to_insert['values'][i] for i in order)
+        return row_to_insert
+
 
 def lookup_class_name(name, context, depth=3):
     """
@@ -679,29 +742,21 @@ class FreeTable(Table):
     """
     A base relation without a dedicated class. Each instance is associated with a table
     specified by full_table_name.
-    :param arg:  a dj.Connection or a dj.FreeTable
+    :param conn:  a dj.Connection object
+    :param full_table_name: in format `database`.`table_name`
     """
-
-    def __init__(self, arg, full_table_name=None):
-        super().__init__()
-        if isinstance(arg, FreeTable):
-            # copy constructor
-            self.database = arg.database
-            self._table_name = arg._table_name
-            self._connection = arg._connection
-        else:
-            self.database, self._table_name = (s.strip('`') for s in full_table_name.split('.'))
-            self._connection = arg
+    def __init__(self, conn, full_table_name):
+        self.database, self._table_name = (s.strip('`') for s in full_table_name.split('.'))
+        self._connection = conn
+        self._support = [full_table_name]
+        self._heading = Heading(table_info=dict(
+            conn=conn,
+            database=self.database,
+            table_name=self.table_name,
+            context=None))
 
     def __repr__(self):
-        return "FreeTable(`%s`.`%s`)" % (self.database, self._table_name)
-
-    @property
-    def table_name(self):
-        """
-        :return: the table name in the schema
-        """
-        return self._table_name
+        return "FreeTable(`%s`.`%s`)\n" % (self.database, self._table_name) + super().__repr__()
 
 
 class Log(Table):
@@ -711,21 +766,20 @@ class Log(Table):
     :param skip_logging: if True, then log entry is skipped by default. See __call__
     """
 
-    def __init__(self, arg, database=None, skip_logging=False):
-        super().__init__()
-
-        if isinstance(arg, Log):
-            # copy constructor
-            self.database = arg.database
-            self.skip_logging = arg.skip_logging
-            self._connection = arg._connection
-            self._definition = arg._definition
-            self._user = arg._user
-            return
+    _table_name = '~log'
 
+    def __init__(self, conn, database, skip_logging=False):
         self.database = database
         self.skip_logging = skip_logging
-        self._connection = arg
+        self._connection = conn
+        self._heading = Heading(table_info=dict(
+            conn=conn,
+            database=database,
+            table_name=self.table_name,
+            context=None
+        ))
+        self._support = [self.full_table_name]
+
         self._definition = """    # event logging table for `{database}`
         id       :int unsigned auto_increment     # event order id
         ---
@@ -736,6 +790,8 @@ class Log(Table):
         event="" :varchar(255)                    # event message
         """.format(database=database)
 
+        super().__init__()
+
         if not self.is_declared:
             self.declare()
             self.connection.dependencies.clear()
@@ -745,10 +801,6 @@ class Log(Table):
     def definition(self):
         return self._definition
 
-    @property
-    def table_name(self):
-        return '~log'
-
     def __call__(self, event, skip_logging=None):
         """
         :param event: string to write into the log table
